(*  
 *  This file is a component of PropPlan a model-based planner
 *  Copyright (C) 2000-2007  Michael Paul Fourman

 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.

 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)
(*****************************************************************************************)
(**** $Id: PDDLSIG.ML,v 1.1 2007/12/30 15:31:25 michaelfourman Exp $ ****)
(*****************************************************************************************)
infix 5 -- ==> ;
infix 3 >>;
infix 0 ||;

infix 4 &&;
signature PDDLSIG =
    sig
	include PROPOSITIONSIG
	type instream
	type token
	exception SynError of string * token list
	type AtomicFormulaSkeleton

	datatype Effect = 
	    EffAtom of Formula                            (* includes simple quantified effects *)
	  | EffWhen of Formula * Formula
	  | EffForall of Term TypedList * (Formula  * Formula)          (* quantified when *)
	  | EffAnd of Effect list

	datatype Action = 
	    Action of {
		       name: Name, 
		       parameters: Term TypedList,
		       precondition: Formula,
		       effect: Effect
		       }

	val actEffect : Action -> Effect
	val predicatesInEffect : Effect -> Predicate list

	datatype Domain = 
	    Domain of {axiom:Formula,
		       name: Name,
		       types: Name TypedList, 
		       constants: Term TypedList, 
		       predicates: AtomicFormulaSkeleton list, 
		       require: Name list,
		       actions: Action list
		       }

	datatype Problem =
	    Problem of {
			name: Name, 
			domain: Name,
			objects: Term TypedList, 
			init: Formula,
			goal: Formula}

	datatype Batch = 
	    Batch of {
		      name : Name, 
		      types: Name TypedList, 
		      actions: Action list,
		      axiom: Formula,
		      require: Name list, 
		      predicates: AtomicFormulaSkeleton list,
		      objects: Term TypedList, 
		      init: Formula,
		      goals: (Name * Formula) list
		      }
	val mkBatch : Domain * Problem list -> Batch list

	val readBatch  : instream -> Batch list
	val readDomain : instream -> Domain
	val readProblem: instream -> Problem

    end;
    
    
    
    
